/*
	IDropTarget Implementation for Drag and Drop

	Copyright (c) 2002 Joshua Seagoe
	Copyright (c) 2004 Sergey Gagarin a.k.a. Seg@
 
	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.
 
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
 
	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "stdafx.h"
#include "droptarget.h"
#include "utils.h"
#include "wharf.h"

using namespace utils;

UINT WharfDropTarget::cf_types[] =
{
	CF_HDROP,
	0,					// CFSTR_INETURL
	0,					// CFSTR_SHELLURL
	CF_TEXT,
	CF_UNICODETEXT,
#ifdef USE_PIDL
	0,					// CFSTR_SHELLIDLIST
#endif
	0
};

/* It's like a switch statement, only greasier! */ 
WharfDropTarget::DropFuncPtr WharfDropTarget::cf_actions[] =
{
	&WharfDropTarget::Drop_HDROP,
	&WharfDropTarget::Drop_INETURL,
	&WharfDropTarget::Drop_SHELLURL,
	&WharfDropTarget::Drop_TEXT,
	&WharfDropTarget::Drop_UNICODETEXT,
#ifdef USE_PIDL
	&WharfDropTarget::Drop_SHELLIDLIST,
#endif
	0
};

void WharfDropTarget::RegisterFormats(void)
{
	// Why did Microsoft think it would be a good idea to force people to do this?
	cf_types[1] = RegisterClipboardFormat(CFSTR_INETURL);
	cf_types[2] = RegisterClipboardFormat(CFSTR_SHELLURL);

#ifdef USE_PIDL
	cf_types[5] = RegisterClipboardFormat(CFSTR_SHELLIDLIST);
#endif
}


HRESULT STDMETHODCALLTYPE WharfDropTarget::QueryInterface( 
	REFIID riid,
	void __RPC_FAR *__RPC_FAR *ppvObject)
{
	return E_NOINTERFACE;
}

ULONG STDMETHODCALLTYPE WharfDropTarget::AddRef( void)
{
	return 1;
}

ULONG STDMETHODCALLTYPE WharfDropTarget::Release( void)
{
	return 1;
}

HRESULT STDMETHODCALLTYPE WharfDropTarget::DragEnter(
	IDataObject *pDataObj,
	DWORD grfKeyState,
	POINTL pt,
	DWORD *pdwEffect)
{
	FORMATETC etc;
	ZeroMemory(&etc,sizeof(FORMATETC));

	m_pWharf->SetDropHover();

	/* We want all of the content, in a global handle. */ 
	etc.tymed = TYMED_HGLOBAL;
	etc.dwAspect = DVASPECT_CONTENT;
	etc.lindex = -1;

	cf_index = 0;
	while (cf_types[cf_index] != 0)
	{
		etc.cfFormat = cf_types[cf_index];
		if (pDataObj->QueryGetData(&etc) == S_OK)
		{
			*pdwEffect = DROPEFFECT_COPY;
			return S_OK;
		}
		++cf_index;
	}
	*pdwEffect = DROPEFFECT_NONE;
	return S_OK;
}

HRESULT STDMETHODCALLTYPE WharfDropTarget::DragOver( 
	DWORD grfKeyState,
	POINTL pt,
	DWORD *pdwEffect)
{
	m_pWharf->SetDropHover();

	// forward it to the window to get hovering
	// mouse messages don't get sent to the main window during drag&drop

	//SendMessage(m_pWharf->m_hWnd, WM_MOUSEMOVE, (WPARAM)grfKeyState, MAKELONG(pt.x,pt.y));

	return S_OK;
}

HRESULT STDMETHODCALLTYPE WharfDropTarget::DragLeave( void)
{
	m_pWharf->ClearDropHover();

	return S_OK;
}

HRESULT STDMETHODCALLTYPE WharfDropTarget::Drop( 
	IDataObject *pDataObj,
	DWORD grfKeyState,
	POINTL pt,
	DWORD *pdwEffect)
{
	FORMATETC etc;
	STGMEDIUM stg;
	ZeroMemory(&etc,sizeof(FORMATETC));

	m_pWharf->ClearDropHover();

	/* We want all of the content, in a global handle. */ 
	etc.tymed = TYMED_HGLOBAL;
	etc.dwAspect = DVASPECT_CONTENT;
	etc.lindex = -1;

	etc.cfFormat = cf_index;
	if (pDataObj->QueryGetData(&etc) != S_OK)
	{
		cf_index = 0;
		while (cf_types[cf_index] != 0)
		{
			etc.cfFormat = cf_types[cf_index];
			if (pDataObj->QueryGetData(&etc) == S_OK) break;
			++cf_index;
		}
		// return if we can't process any data types provided
		if (cf_types[cf_index] == 0)
		{
			return S_OK;
		}
	}

	/* Grab the stuff */ 
	if (pDataObj->GetData(&etc,&stg)==S_OK)
	{
		/* Use it, locked so we get a memory reference. */ 
		(this->*cf_actions[cf_index])(GlobalLock(stg.hGlobal));
		GlobalUnlock(stg.hGlobal);

		/* Free the medium. */ 
		if (stg.pUnkForRelease==NULL)
			GlobalFree(stg.hGlobal);
		else
			stg.pUnkForRelease->Release();
	}

	return S_OK;
}

#define HIDA_GetPIDLParent(pida) (LPCITEMIDLIST)(((LPBYTE)pida)+(pida)->aoffset[0])
#define HIDA_GetPIDLItem(pida, i) (LPCITEMIDLIST)(((LPBYTE)pida)+(pida)->aoffset[i+1])

#ifdef USE_PIDL

void DumpPIDL(LPCITEMIDLIST pidl)
{
#if !defined(NDEBUG) && defined(TRACE_FILE)
	char buffer[MAX_PATH];
	int j;
#endif
#if !defined(NDEBUG) && defined(TRACE_FILE)
		sprintf(buffer,"PIDL:%08X",pidl);
		TRACE(TRACE_DUMP,buffer);

		j = ((SHITEMID *)(pidl))->cb;
		buffer[j+j] = 0;
		while (j--)
		{
			/* convert to hex (scary lookin, innit?) */ 
			buffer[j+j+1] = (reinterpret_cast<const SHITEMID *>(pidl)->abID[j]&0x0f)["0123456789ABCDEF"];
			buffer[j+j+0] = (reinterpret_cast<const SHITEMID *>(pidl)->abID[j] >> 4)["0123456789ABCDEF"];
		}
		TRACE(TRACE_DUMP,buffer);

		j = (reinterpret_cast<const SHITEMID *>(pidl))->cb;
		buffer[j] = 0;
		while (j--)
		{
			buffer[j] = (reinterpret_cast<const SHITEMID *>(pidl)->abID[j]);
			if (buffer[j]<32 || buffer[j]>126) buffer[j] = '-';
		}
		TRACE(TRACE_DUMP,buffer);

		SHGetPathFromIDList(reinterpret_cast<LPCITEMIDLIST>(pidl),buffer);
		TRACE(TRACE_DUMP,buffer);
#endif
}

UINT GetSizePIDL(LPCITEMIDLIST pidl)
{
	UINT i = (pidl->mkid.cb);

	if (i==0) return sizeof(pidl->mkid.cb);

	DumpPIDL(pidl);

	// walk the id list chain
	return i + GetSizePIDL(
		reinterpret_cast<LPCITEMIDLIST>(
		reinterpret_cast<const BYTE *>(pidl) + i
		));
}

LPITEMIDLIST AppendPIDL(LPMALLOC pMalloc,LPCITEMIDLIST pidlBase, LPCITEMIDLIST pidlAdd)
{
    if(pidlBase == NULL)
        return NULL;
    if(pidlAdd == NULL)
        return const_cast<LPITEMIDLIST>(pidlBase);
    
    LPITEMIDLIST pidlNew;

	/*
		Microsoft has an example for this, but it only works
		if the id lists contain only one element each.
		(either that, or they do exactly the same thing as
		this, but don't tell you how to get GetSizePIDL)
	*/

    UINT cb1 = GetSizePIDL(pidlBase);
    UINT cb2 = GetSizePIDL(pidlAdd);

    pidlNew = (LPITEMIDLIST)pMalloc->Alloc(cb1 + cb2);
    if (pidlNew)
    {
        CopyMemory(((LPSTR)pidlNew), pidlBase, cb1);
        CopyMemory(((LPSTR)pidlNew) + cb1 - sizeof(pidlBase->mkid.cb), pidlAdd, cb2);

		DumpPIDL(pidlNew);
    }
    return pidlNew;
}

#endif /* USE_PIDL */ 


#ifdef USE_PIDL

void WharfDropTarget::Drop_SHELLIDLIST(void *files)
{
	SHELLEXECUTEINFO ExecInfo;
	CIDA *pCIDA = reinterpret_cast<CIDA *>(files);
	UINT i = 0;

	ZeroMemory(&ExecInfo,sizeof(ExecInfo));
	ExecInfo.cbSize = sizeof(ExecInfo);
	ExecInfo.nShow = SW_SHOWNORMAL;
	ExecInfo.hwnd = m_pWharf->m_hWnd;

	LPMALLOC pMalloc;

	ExecInfo.fMask = SEE_MASK_INVOKEIDLIST | SEE_MASK_IDLIST | SEE_MASK_FLAG_NO_UI;

	SHGetMalloc(&pMalloc);

	while (i < pCIDA->cidl)
	{
		// stupid shell gives me a folder and a relative pidl
		// where i really need an absolute pidl
		ExecInfo.lpIDList = AppendPIDL(pMalloc,
			HIDA_GetPIDLParent(pCIDA),HIDA_GetPIDLItem(pCIDA,i));

		ShellExecuteEx(&ExecInfo);

		pMalloc->Free(ExecInfo.lpIDList);

		++i;
	}

	pMalloc->Release();
}

#endif /* USE_PIDL */


void WharfDropTarget::Drop_HDROP(void *files)
{
	DROPFILES *df = reinterpret_cast<DROPFILES *>(files);
	wchar_t *names;
	char name[MAX_PATH_LENGTH];

	// damn unicode
	names = reinterpret_cast<wchar_t *>(reinterpret_cast<BYTE *>(files)+df->pFiles);
	while (*names!=0)
	{
		WideCharToMultiByte(CP_ACP,0,names,-1,name,MAX_PATH_LENGTH,NULL,NULL);

		m_pWharf->DropFile(name);

		while (*++names!=0) ;
		++names;
	}
}

void WharfDropTarget::Drop_INETURL(void *text)
{
	char *p = reinterpret_cast<char *>(text);
	char name[MAX_PATH_LENGTH]={0};

	if (p[1]==0)
	{
		// assume urls are never 1 byte long
		// so this one's in unicode
		WideCharToMultiByte(CP_ACP,0,reinterpret_cast<wchar_t *>(text),-1,name,MAX_PATH_LENGTH,NULL,NULL);
		p = name;
	}
	else
	{
		p = reinterpret_cast<char *>(text);
	}

	m_pWharf->DropFile(p);
}

void WharfDropTarget::Drop_SHELLURL(void *text)
{
	Drop_INETURL(text);
}

void WharfDropTarget::Drop_TEXT(void *text)
{
	char *p = reinterpret_cast<char *>(text);
	char *eol;
	char nl = 13;

	// parse it line-for-line and execute it
	
	eol = strchr(p,nl);
	if (!eol)
	{
		nl=10;
		eol = strchr(p,nl);
	}
	while (eol)
	{
		*eol = 0;

		m_pWharf->DropFile(p);

		p = eol+1;
		while (*p==13 || *p==10) ++p;

		eol = strchr(p,nl);
	}
	m_pWharf->DropFile(p);
}

void WharfDropTarget::Drop_UNICODETEXT(void *text)
{
	char buffer[MAX_LINE_LENGTH];

	WideCharToMultiByte(CP_ACP,0,reinterpret_cast<wchar_t *>(text),-1,buffer,
		MAX_LINE_LENGTH,NULL,NULL);

	Drop_TEXT(buffer);
}


HRESULT WharfDropTarget::Register(wharfType *parent)
{
	cf_index=0;
	m_pWharf = parent;
	return RegisterDragDrop(m_pWharf->m_hWnd, this);
}