/*

This is a part of the LiteStep Shell Source code.
 
Copyright (c) 1997-98 The LiteStep Development Team
Copyright (c) 2004 Sergey Gagarin a.k.a. Seg@

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
	
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
	  
*/

#include "stdafx.h"
#include "wharf.h"
#include "utils.h"
#include "droptarget.h"

using namespace utils;

// -------------------------------------------------------------------------------------------------------

const char szMainWindowClass[] = "TWharfGlobalContainer";
const char szHiddenWindowClass[] = "LineDownSide";
const char rcsRevision[] = "2.1.0"; // Our Version


// our window procedures
LRESULT CALLBACK MainWndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK HiddenWndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);

// -------------------------------------------------------------------------------------------------------

void CloseAllFolders();
bool PartOfWharf(HWND myWnd);
void ReadConfig(void);
int GetWharfByWnd(HWND hwnd);
void SetWorkArea(void);
void ShowWharf(void);
void HideWharf(void);
void CreateWharfhints(HWND hWnd, char *txt, RECT *r);
void RemoveWharfhints(HWND hWnd);
void SoundOpen(void);
void SoundClose(void);
void SoundMin(void);
void SoundMax(void);
void BangWharfRollup(HWND hCaller, LPCSTR pszArgs);
void BangWharfShow(HWND hCaller, LPCSTR pszArgs);
void BangWharfHide(HWND hCaller, LPCSTR pszArgs);

char szLitestepPath[MAX_LINE_LENGTH];
char lsPath[MAX_LINE_LENGTH];

char wharfOpenSound[MAX_LINE_LENGTH];
char wharfCloseSound[MAX_LINE_LENGTH];
char wharfMinSound[MAX_LINE_LENGTH];
char wharfMaxSound[MAX_LINE_LENGTH];

char szTitleBarLeftClick[MAX_LINE_LENGTH];
char szTitleBarRightClick[MAX_LINE_LENGTH];

char szImagePath[MAX_LINE_LENGTH];
char szDefPixmap[MAX_LINE_LENGTH];
char szFolderPixmap[MAX_LINE_LENGTH];
HBITMAP titlebarImage = NULL;
HBITMAP defaultBackImage = NULL;
HBITMAP folderImage = NULL;
HBITMAP defaultFolderImage = NULL;
HINSTANCE dll;
HWND hMainWnd;
HWND HiddenWnd;
HWND hwndLitestep;
HWND oldCapt = NULL;
HWND wharfHints = NULL;
HRGN hMainRgn;
int ScreenWidth, ScreenHeight, wharfTileSize;
int wharfBaseX = 0, wharfBaseY = 0;

WharfVector wharfs;

int wharfDirection = WHARF_DOWN;

int titleSize;
int autoHideDelay = 300;
int autoShowDelay = 300;
int hiddenWidth = 1;
int OpenFolders = 0;
int wharfShadeButton;
int wharfDblClickXPosition = 0;
bool wharfDblClickDockOnLeft = false;
int wharfClosingPos;
int wharfAnimTime;
int wharfDocked = 0;
int wharfStep = 64;
int wharfPressOffset = 1;
bool wharfAlwaysOnTop = true;
bool wharfSlidersOnTop = true;
bool wharfClosing = false;
bool wharfHide = false;
bool wharfTitlebar = true;
bool wharfNoAnim = false;
bool wharfAutoClose = true;
bool wharfTitles = false;
bool wharfCloseOnSwitch = false;
bool snapTo = false;
bool wharfAutoHide = false;
bool hideTimerActive = false;
bool showTimerActive = false;
bool wharfHidden = false;
bool wharfAutoUnpress = false;
bool wharfNoHints = true;
int nIconSize;
int snapToSensitivity = 16;
int bevelWidth;
int capHeight;
COLORREF titleFore, titleBack;
char szTitleFont[MAX_LINE_LENGTH] = "Arial";
int iTitleFontSize = 8, iTitleFontWeight = 400;
bool setDesktopArea = true;

bool isFirst = true;
char szModulesINI[MAX_LINE_LENGTH] = "";

const UINT Msgs[] = {LM_SHADETOGGLE, LM_GETREVID, 0};

#define RETURN_BOX(msg,rv) {ErrorBox(msg);return rv;}

int GetWharfByPointer(wharfType *w)
{
	int i;

	for (i = 0;i < wharfs.size();i++)
	{
		if (wharfs[i] == w) return i;
	}
	return -1;
}

int GetWharfByWnd(HWND hwnd)
{
	int i;

	for (i = 0;i < wharfs.size();i++)
	{
		if (wharfs[i]->m_hWnd == hwnd) return i;
	}
	return -1;
}

// -------------------------------------------------------------------------------------------------------
int initModuleEx(HWND ParentWnd, HINSTANCE dllInst, LPCSTR szPath)
{
	
	RECT r;
	int i;
	HWND desktop;

	dll = dllInst;

	GetClientRect(GetDesktopWindow(), &r);
	ScreenWidth	= GetSystemMetrics(SM_CXSCREEN);
	ScreenHeight= GetSystemMetrics(SM_CYSCREEN);

	hwndLitestep = GetLitestepWnd();

	StringCchCopy(szLitestepPath, MAX_LINE_LENGTH, szPath);

	ReadConfig();

	desktop = FindWindow("DesktopBackgroundClass", NULL);
	if (!desktop)
	{
		desktop = GetDesktopWindow();
	}

	// create hints window
	if (!wharfNoHints)
	{
		wharfHints = CreateWindowEx(WS_EX_TOOLWINDOW | WS_EX_TOPMOST,
			TOOLTIPS_CLASS,
			"LSWharfHints",
			TTS_ALWAYSTIP,
			CW_USEDEFAULT,
			CW_USEDEFAULT,
			CW_USEDEFAULT,
			CW_USEDEFAULT,
			NULL,
			(HMENU) NULL,
			dll,
			NULL
			);
		if (!wharfHints) RETURN_BOX("Error creating hints window", 1);
	}

	titleSize = wharfTitlebar ? capHeight : 0;

	{	// Register our window class
		WNDCLASS wc;

		ZeroMemory(&wc, sizeof(wc));
		wc.lpfnWndProc = MainWndProc;				  // our window procedure
		wc.hInstance = dllInst; 				// hInstance of DLL
		wc.lpszClassName = szMainWindowClass;	   // our window class name
		wc.style = CS_DBLCLKS;

		if (!RegisterClass(&wc)) RETURN_BOX("Error registering window class", 1);

		if (wharfAutoHide)
		{
			WNDCLASS wc;
			ZeroMemory(&wc, sizeof(wc));
			wc.lpfnWndProc = HiddenWndProc;				// our window procedure
			wc.hInstance = dllInst;						// hInstance of DLL
			wc.lpszClassName = szHiddenWindowClass;		// our window class name
			wc.style = 0;

			if (!RegisterClass(&wc)) RETURN_BOX("Error registering window class", 1);
		}
	}
	hMainWnd = CreateWindowEx(
	               WS_EX_TOOLWINDOW | (wharfAlwaysOnTop ? WS_EX_TOPMOST : 0),
	               szMainWindowClass,
	               "LSWharf",
	               WS_CLIPCHILDREN | WS_POPUP,
	               wharfBaseX,
	               wharfBaseY,
	               0,
	               0,
				   wharfAlwaysOnTop ? NULL : desktop,
	               NULL,
	               dllInst,
	               NULL
	           );

	if (!hMainWnd) RETURN_BOX("Error creating wharf", 1);

	SetWindowLong (hMainWnd, GWL_USERDATA, magicDWord);

	if (wharfAutoHide)
	{
		HiddenWnd = CreateWindowEx(
		                WS_EX_TOPMOST | WS_EX_TOOLWINDOW,
		                szHiddenWindowClass,
		                "LSWharfHidden",
		                WS_POPUP,
		                (wharfBaseX == 0) ? 0 : wharfBaseX + wharfTileSize - hiddenWidth,
		                wharfBaseY,
		                hiddenWidth,
		                ScreenHeight,
		                NULL,
		                NULL,
		                dllInst,
		                NULL
		            );

		if (!HiddenWnd) RETURN_BOX("Error creating wharf", 1);

		SetWindowLong (HiddenWnd, GWL_USERDATA, magicDWord);
	}

	WharfDropTarget::RegisterFormats();

	{
		FILE *f;

		f = LCOpen (NULL);
		if (f)
		{
			char	buffer[MAX_LINE_LENGTH];
			char
				token1[MAX_LINE_LENGTH], token2[MAX_LINE_LENGTH],
				token3[MAX_LINE_LENGTH], token4[MAX_LINE_LENGTH], extra_text[MAX_LINE_LENGTH];
			char*	tokens[4] = {token1, token2, token3, token4};
			bool inFolder = false;

			buffer[0] = 0;

			while (LCReadNextConfig (f, "*Wharf", buffer, sizeof (buffer)))
			{
				int count;

				token1[0] = token2[0] = token3[0] = token4[0] = extra_text[0] = '\0';

				count = LCTokenize (buffer, tokens, 4, extra_text);

				switch (count)
				{
				case 2:
					if (!strcmpi(token2, "~Folder"))
					{
						inFolder = FALSE;
						break;
					}
				case 4:
				{
					// token 1 is "*Wharf", token 2 is the name, token 3 is the
					// graphic, token 4 is the executable, and the extra_text are
					// the program arguments.
					char windowName[MAX_LINE_LENGTH] = "\0";

					if ((token4[0] == '%') && (strrchr(token4, '%') == token4))
					{
						char szTempToken[MAX_LINE_LENGTH] = {0};
						char*	tokens2[5] = {token1, token2, token3, szTempToken, token4};
						StringCchCopy(windowName, MAX_LINE_LENGTH, token4 + 1);
						extra_text[0] = token4[0] = '\0';
					}
					if (inFolder)
					{
						char szTT[MAX_LINE_LENGTH];
						int w = wharfs.size() - 1;
						subwharfType *shortcut = new subwharfType(wharfs[w]);
						wharfs[w]->subwharfs.push_back(shortcut);
						if (token3[0] != '*')
							shortcut->m_bShowBackground = true;
						else
						{
							char *temptoken3;
							temptoken3 = token3;
							if (*temptoken3 == '*')
							{
								temptoken3++;
							}
							StringCchCopy(token3, MAX_LINE_LENGTH, temptoken3);
							shortcut->m_bShowBackground = false;
						}

						StringCchCopy(shortcut->m_szName, MAX_LINE_LENGTH, token2);
						StringCchCopy(shortcut->m_szCommand, MAX_LINE_LENGTH, token4);
						StringCchCopy(shortcut->m_szParameters, MAX_LINE_LENGTH, extra_text);

						StringCchCopy(szTT, MAX_LINE_LENGTH, token4);
						if (strnicmp(".extract", token3, 8) && extra_text && strlen(extra_text))
						{
							StringCchCat (szTT, MAX_LINE_LENGTH, " ");
							StringCchCat (szTT, MAX_LINE_LENGTH, extra_text);
						}

						shortcut->LoadImage (token3, szTT);

						shortcut->Create(windowName);
					}
					else
					{
						char szTT[MAX_LINE_LENGTH];

						wharfType *shortcut = new wharfType();
						wharfs.push_back(shortcut);
						if (token3[0] == '*')
						{
							char *temptoken3;
							temptoken3 = token3;
							if (*temptoken3 == '*') temptoken3++;
							StringCchCopy(token3, MAX_LINE_LENGTH, temptoken3);
							shortcut->m_bShowBackground = false;
						}
						else
						{
							shortcut->m_bShowBackground = true;
						}

						StringCchCopy(shortcut->m_szName, MAX_LINE_LENGTH, token2);
						StringCchCopy(shortcut->m_szCommand, MAX_LINE_LENGTH, token4);
						StringCchCopy(shortcut->m_szParameters, MAX_LINE_LENGTH, extra_text);

						if (!strcmpi (token4, "Folder"))
						{
							StringCchCopy(szTT, MAX_LINE_LENGTH, token2);
						}
						else
						{
							StringCchCopy(szTT, MAX_LINE_LENGTH, token4);
							if (strnicmp(".extract", token3, 8) && extra_text && strlen(extra_text))
							{
								StringCchCat (szTT, MAX_LINE_LENGTH, " ");
								StringCchCat (szTT, MAX_LINE_LENGTH, extra_text);
							}
						}
						shortcut->LoadImage(token3, szTT);

						shortcut->Create(windowName);

						if (!strcmpi(token4, "Folder"))
						{
							inFolder = true;
							shortcut->m_bSubwharfOpen = false;
							shortcut->m_bPressed = false;
							shortcut->m_bTogglePressed = false;
							shortcut->m_hwndSubwharf = CreateWindowEx(
								(wharfSlidersOnTop ? WS_EX_TOPMOST : 0) | WS_EX_TOOLWINDOW,
								szMainWindowClass,
								"LSWharflet",
								WS_CLIPCHILDREN | WS_POPUP,
								0, 0,
								0, 0,
								wharfSlidersOnTop ? NULL : hwndLitestep,
								NULL,
								dll,
								NULL
								);

							SetWindowLong(shortcut->m_hwndSubwharf, GWL_USERDATA, magicDWord);
							ShowWindow(shortcut->m_hwndSubwharf, SW_SHOWNORMAL);
						}
					}
				}
				default:
					break;
				}
			}
			LCClose(f);
		}
	}

	SendMessage(hwndLitestep, LM_REGISTERMESSAGE, (WPARAM) hMainWnd, (LPARAM) Msgs);

	/* transparent titlebar */
	if (wharfTitlebar && titlebarImage)
	{
		HRGN hClipRgn, hTitlebarRgn;

		hClipRgn = CreateRectRgn(0, 0, wharfTileSize, capHeight);
		hTitlebarRgn = BitmapToRegion(titlebarImage, RGB(255, 0, 255), 0x101010, 0, 0);

		if (hMainRgn)
		{
			CombineRgn(hTitlebarRgn, hClipRgn, hTitlebarRgn, RGN_DIFF);
			CombineRgn(hMainRgn, hMainRgn, hTitlebarRgn, RGN_DIFF);
		}

		DeleteObject(hClipRgn);
		DeleteObject(hTitlebarRgn);
	}

	/* set our region for hMainWnd now, this sets the wharf into our transparency mode     */
	/* after we compiled the full region for the parent window of the child, tile, windows */
	/* C. Boyda                                                                            */

	SetWindowRgn(hMainWnd, hMainRgn, TRUE);
	hMainRgn = NULL;

	for (i = 0; i < wharfs.size(); i++)
	{
		if (wharfs[i]->m_hwndSubwharf)
		{
			SetWindowRgn(wharfs[i]->m_hwndSubwharf, wharfs[i]->m_hrgnSubRgn, TRUE);
			wharfs[i]->m_hrgnSubRgn = NULL;
		}
	}

	if (!wharfHidden)
    {
        // If we are autohiding, set the autohide delay.
        if (wharfAutoHide)
        {
            RECT r;
            GetClientRect(hMainWnd, &r);
            
            SetWindowPos(HiddenWnd, 0, 0, 0, hiddenWidth, r.bottom - r.top, SWP_NOZORDER | SWP_NOMOVE);
            
            SetTimer(HiddenWnd, 1, 50, NULL);
        }
    }
    else
    {
        SetWindowPos(hMainWnd, 0, wharfBaseX, wharfBaseY, wharfTileSize, titleSize, SWP_NOZORDER);
    }

    ShowWindow(hMainWnd, SW_SHOWNORMAL);
    
	SetWorkArea();
	isFirst = false;

	AddBangCommand("!WharfRollup", BangWharfRollup);
    AddBangCommand("!WharfShow", BangWharfShow);
    AddBangCommand("!WharfHide", BangWharfHide);
	
	return 0;
}

// -------------------------------------------------------------------------------------------------------

void HiddenWnd_DisplayChange(Message& message)
{
	bool WasDocked = (wharfBaseX + wharfTileSize) == ScreenWidth;
	int RelativeX = ScreenWidth - wharfBaseX;
	bool MoveRelative = RelativeX > (ScreenWidth / 2);
	CloseAllFolders();
	ScreenWidth = GetSystemMetrics(SM_CXSCREEN);
	ScreenHeight = GetSystemMetrics(SM_CYSCREEN);
	if (WasDocked)
		wharfBaseX = ScreenWidth - wharfTileSize;
	else if (MoveRelative)
		wharfBaseX = ScreenWidth - RelativeX;
	SetWindowPos(hMainWnd, HWND_TOPMOST, wharfBaseX, wharfBaseY, 0, 0, SWP_NOZORDER | SWP_NOSIZE | SWP_NOACTIVATE);
}


void HiddenWnd_Paint(Message& message)
{
	PAINTSTRUCT ps;
	RECT r;
	HDC hdc = BeginPaint(HiddenWnd, &ps);
	HBRUSH hb = CreateSolidBrush(0xFFFFFF);

	GetClientRect(HiddenWnd, &r);
	r.right += 1;
	r.bottom += 1;

	FillRect(hdc, &r, hb);

	DeleteObject(hb);
	EndPaint(HiddenWnd, &ps);
}


void HiddenWnd_Timer(Message& message)
{
	if (wharfDocked)
	{
		switch (message.wParam)
		{
		case 0:
			if (wharfAutoHide)
			{
				HideWharf();
			}
			KillTimer(HiddenWnd, 0);
			hideTimerActive = FALSE;
			break;
		case 1:
			if (wharfAutoHide && !wharfHidden && !hideTimerActive)
			{
				POINT pt;
				RECT r;
				GetCursorPos(&pt);
				GetWindowRect(hMainWnd, &r);
				if ((pt.x < r.left || pt.x > r.right || pt.y < r.top || pt.y > r.bottom) &&
				        (wharfNoAnim || wharfCloseOnSwitch ? TRUE : !OpenFolders))
				{
					SetTimer(HiddenWnd, 0, autoHideDelay, NULL);
					hideTimerActive = TRUE;
				}
			}
			else
			{
				if (wharfAutoHide && wharfHidden && !showTimerActive)
				{
					POINT pt;
					RECT r;
					GetCursorPos(&pt);
					GetWindowRect(HiddenWnd, &r);
					if (pt.x >= r.left && pt.x <= r.right && pt.y >= r.top && pt.y <= r.bottom)
					{
						SetTimer(HiddenWnd, 2, autoShowDelay, NULL);
						showTimerActive = TRUE;
					}

				}
				else if (wharfHidden && showTimerActive)
				{
					POINT pt;
					RECT r;
					GetCursorPos(&pt);
					GetWindowRect(HiddenWnd, &r);
					if (pt.x < r.left || pt.x > r.right || pt.y < r.top || pt.y > r.bottom)
					{
						KillTimer(HiddenWnd, 2);
						showTimerActive = FALSE;
					}
				}
			}
			break;
		case 2:
			if (wharfAutoHide)
			{
				ShowWharf();
			}
			KillTimer(HiddenWnd, 2);
			showTimerActive = FALSE;
			break;

		}
	}
	else
	{
		if (hideTimerActive)
		{
			KillTimer(HiddenWnd, 0);
			hideTimerActive = FALSE;
		}
	}
}


LRESULT CALLBACK HiddenWndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	Message message;

	message.uMsg = uMsg;
	message.wParam = wParam;
	message.lParam = lParam;
	message.lResult = 0;

	switch (uMsg)
	{
	case WM_ENDSESSION:
	case WM_QUERYENDSESSION:
		return SendMessage(hwndLitestep, message.uMsg, message.wParam, message.lParam);
	case WM_SYSCOMMAND:
		if (message.wParam == SC_CLOSE)
		{
			PostMessage(hwndLitestep, WM_KEYDOWN, LM_SHUTDOWN, 0);
		}
		break;
	case WM_DISPLAYCHANGE:
		HiddenWnd_DisplayChange(message);
		break;
	case WM_PAINT:
		HiddenWnd_Paint(message);
		break;
	case WM_TIMER:
		HiddenWnd_Timer(message);
		break;
	default:
		return DefWindowProc(hWnd, uMsg, wParam, lParam);
	}
	return message.lResult;
}


void MainWnd_GetRevId(Message& message)
{
	char* buf = (char*)(message.lParam);

	switch (message.wParam)
	{
	case 0:
	case 1:
		StringCchCopy(buf, MAX_LINE_LENGTH, "wharf.dll: ");
		StringCchCat(buf, MAX_LINE_LENGTH, rcsRevision);
		message.lResult = strlen(buf);
		break;
	default:
		StringCchCopy(buf, MAX_LINE_LENGTH, "");
	}
}


void MainWnd_ShadeToggle(Message& message)
{
	if (wharfClosing) return ;
	CloseAllFolders();
	if (!wharfHide)
	{
		if (wharfMinSound[0]) SoundMin();
		if (wharfNoAnim)
		{
			SetWindowPos(hMainWnd, 0, wharfBaseX, wharfBaseY, wharfTileSize, titleSize, SWP_NOZORDER);
			wharfHide = TRUE;
			return;
		}
		wharfClosing = true;
		wharfClosingPos = (wharfs.size() * wharfTileSize) + titleSize;
		SetTimer(hMainWnd, 1, wharfAnimTime, NULL);
	}
	else
	{
		if (wharfMaxSound[0]) SoundMax();
		if (wharfNoAnim)
		{
			SetWindowPos(
				hMainWnd,
				0,
				wharfBaseX,
				wharfBaseY,
				wharfTileSize,
				(wharfs.size() * wharfTileSize) + titleSize,
				SWP_NOZORDER);
			wharfHide = FALSE;
			return ;
		}
		wharfClosing = true;
		wharfClosingPos = titleSize;
		SetTimer(hMainWnd, 0, wharfAnimTime, NULL);
	}
}


void MainWnd_Timer(HWND hWnd, Message& message)
{
	if (wharfClosing && message.wParam == 0)
	{
		wharfClosingPos += wharfStep;
		SetWindowPos(hMainWnd, 0, wharfBaseX, wharfBaseY, wharfTileSize, wharfClosingPos, SWP_NOZORDER);
		if (wharfClosingPos >= (wharfs.size() * wharfTileSize) + titleSize)
		{
			wharfHide = FALSE;
			wharfClosing = false;
			KillTimer(hMainWnd, 0);
		}
	}
	if (wharfClosing && message.wParam == 1)
	{
		if (wharfClosingPos - wharfStep <= titleSize )
			wharfClosingPos = titleSize;
		else
			wharfClosingPos -= wharfStep;
		SetWindowPos(hMainWnd, 0, wharfBaseX, wharfBaseY, wharfTileSize, wharfClosingPos, SWP_NOZORDER);
		if (wharfClosingPos <= titleSize)
		{
			wharfHide = TRUE;
			wharfClosing = false;
			KillTimer(hMainWnd, 1);
		}
	}
	if (message.wParam == 2)   //sub wharf opening
	{
		int x, i = -1;

		for (x = 0; x < wharfs.size(); x++)
		{
			if (wharfs[x]->m_hwndSubwharf == hWnd)
			{
				i = x;
				break;
			}
		}

		if (i != -1) // subwharf positions
		{
			wharfs[i]->onTimerOpen();
		}
	}
	if (message.wParam == 3)  // sub wharf closing
	{
		for (int i = 0; i < wharfs.size(); i++)
		{
			if (wharfs[i]->m_hwndSubwharf == hWnd)
			{
				wharfs[i]->onTimerClose();
			}
		}
	}
	if (message.wParam == 4)
	{
		for (int i = 0; i < wharfs.size(); i++)
		{
			POINT pt;
			GetCursorPos(&pt);
			if (wharfs[i]->m_bPressed && !wharfs[i]->m_bSubwharfOpen && !PtInRegion(hMainRgn, pt.x, pt.y))
			{
				wharfs[i]->togglePressOffset(false);
			}
		}

	}
}


void MainWnd_NCButtonDown(HWND hWnd, Message& message)
{
	if ((int)(message.lParamLo) - wharfBaseX >= wharfShadeButton)
	{
		switch (message.uMsg)
        {
            case WM_NCLBUTTONDOWN:
            {
                LSExecute(hWnd, szTitleBarLeftClick, SW_SHOWNORMAL);
                break;
            }
            
            case WM_NCRBUTTONDOWN:
            {
                LSExecute(hWnd, szTitleBarRightClick, SW_SHOWNORMAL);
                break;
            }
        }
	}
	else
		message.lResult = DefWindowProc(hWnd, message.uMsg, message.wParam, message.lParam);
}


void MainWnd_NCLButtonDblClick(HWND hWnd, Message& message)
{
	if ((hWnd == hMainWnd) && wharfTitlebar)
	{
		if (wharfHide)
		{
			wharfBaseX = wharfDblClickDockOnLeft ? 0 : ScreenWidth - wharfTileSize;
		}
		else
		{
			wharfBaseX = ScreenWidth - wharfDblClickXPosition;
			if (wharfBaseX >= ScreenWidth - wharfTileSize)
				wharfBaseX = ScreenWidth - wharfTileSize;
		}
		wharfBaseY = 0;
		SetWorkArea();
		SendMessage(hWnd, LM_SHADETOGGLE, 0, 0);
	}
}


void MainWnd_WindowPosChanging(HWND hWnd, Message& message)
{
	if ((hWnd == hMainWnd) && wharfTitlebar)
	{
		LPWINDOWPOS lpwp = (LPWINDOWPOS)(message.lParam);

		if (lpwp->y <= 0)
			lpwp->y = 0;
		else if (lpwp->y >= ScreenHeight - capHeight)
			lpwp->y = ScreenHeight - capHeight;

		if (lpwp->x >= ScreenWidth - wharfTileSize)
			lpwp->x = ScreenWidth - wharfTileSize;
		else if (lpwp->x <= 0)
			lpwp->x = 0;

		if (snapTo)
			if (lpwp->y <= snapToSensitivity)
				lpwp->y = 0;

		if (lpwp->x >= ScreenWidth - wharfTileSize - snapToSensitivity)
			lpwp->x = ScreenWidth - wharfTileSize;
		else if (lpwp->x <= snapToSensitivity)
			lpwp->x = 0;

	}
	else
		message.lResult = DefWindowProc(hWnd, message.uMsg, message.wParam, message.lParam);
}


void MainWnd_NCHitTest(HWND hWnd, Message& message)
{
	if ((hWnd == hMainWnd) && wharfTitlebar)
		message.lResult = HTCAPTION;
	else
		message.lResult = DefWindowProc(hWnd, message.uMsg, message.wParam, message.lParam);
}


void MainWnd_Move(HWND hWnd, Message& message)
{
	if (hWnd != hMainWnd) return ;

	char szBuf[MAX_LINE_LENGTH];

	CloseAllFolders();

	wharfBaseX = message.lParamLo;
	wharfBaseY = message.lParamHi;

	SetWorkArea();

	StringCchPrintf(szBuf, MAX_LINE_LENGTH, "%d,%d", wharfBaseX, wharfBaseY);
	WritePrivateProfileString("Wharf", "Position", szBuf, szModulesINI);
}


void MainWnd_DisplayChange(Message& message)
{
	bool WasDocked = (wharfBaseX + wharfTileSize) == ScreenWidth;
	int RelativeX = ScreenWidth - wharfBaseX;
	bool MoveRelative = RelativeX > (ScreenWidth / 2);
	CloseAllFolders();
	ScreenWidth = GetSystemMetrics(SM_CXSCREEN);
	ScreenHeight = GetSystemMetrics(SM_CYSCREEN);
	if (WasDocked)
		wharfBaseX = ScreenWidth - wharfTileSize;
	else if (MoveRelative)
		wharfBaseX = ScreenWidth - RelativeX;
	SetWindowPos(hMainWnd, HWND_TOPMOST, wharfBaseX, wharfBaseY, 0, 0, SWP_NOZORDER | SWP_NOSIZE | SWP_NOACTIVATE);
}


void MainWnd_Paint(HWND hWnd, Message& message)
{
	HBITMAP oldBMP = NULL;
	int i = GetWharfByWnd(hWnd);
	PAINTSTRUCT ps;
	HDC hdc = BeginPaint(hWnd, &ps);
	HDC src = CreateCompatibleDC(hdc);

	if (hWnd == hMainWnd && wharfTitlebar && titlebarImage)
	{
		oldBMP = (HBITMAP)SelectObject(src, titlebarImage);
		BitBlt(hdc, 0, 0, wharfTileSize, titleSize, src, 0, 0, SRCCOPY);
		SelectObject(src, oldBMP);
	}

	// if we are painting a wharf...
	if (i != -1)
	{
		wharfs[i]->Paint(hdc, src);
	}
	else
	{
		// not painting a wharf...a subwharf perhaps?

		for (int x = 0; x < wharfs.size(); x++)
		{
			for (int y = 0; y < wharfs[x]->subwharfs.size(); y++)
			{
				// painting this subwharf?
				if (hWnd == wharfs[x]->subwharfs[y]->m_hWnd)
				{
					wharfs[x]->subwharfs[y]->Paint(hdc, src);
				}
			}
		}
	}
	EndPaint(hWnd, &ps);
	DeleteDC(src);
}


void MainWnd_LButtonDown(HWND hWnd, Message& message)
{
	int i = GetWharfByWnd(hWnd);

	for (int x = 0; x < wharfs.size(); x++)
	{
		wharfs[x]->m_bPressed = false;
	}

	if (i != -1)
	{
		wharfs[i]->m_bPressed = true;
		wharfs[i]->togglePressOffset(true);
		SetTimer(wharfs[i]->m_hWnd, 4, 50, NULL);
	}
}

void MainWnd_LButtonUp(HWND hWnd, Message& message)
{
	int i = GetWharfByWnd(hWnd);

	if (i != -1 && wharfs[i]->m_bPressed)
	{
		if (wharfAutoUnpress)
		{
			wharfs[i]->m_bPressed = false;
			wharfs[i]->togglePressOffset(false);
		}

		KillTimer(wharfs[i]->m_hWnd, 4);
		wharfs[i]->Execute();
	}
	else
	{
		for (int x = 0; x < wharfs.size(); x++)
		{
			for (int y = 0; y < wharfs[x]->subwharfs.size(); y++)
			{
				if (hWnd == wharfs[x]->subwharfs[y]->m_hWnd)
				{
					if (wharfAutoClose) CloseAllFolders();
					wharfs[x]->subwharfs[y]->Execute();
					return;
				}
			}
		}
	}
}


void MainWnd_MouseMove(HWND hWnd, Message& message)
{
	MSG TooltipMessage;

	int i = GetWharfByWnd(hWnd);
	if (wharfAutoHide && hideTimerActive)
	{
		KillTimer(HiddenWnd, 0);
		hideTimerActive = FALSE;
	}

	if (!wharfNoHints)
	{
		if (i >= 0)
		{
			TooltipMessage.hwnd = wharfs[i]->m_hWnd;
			TooltipMessage.message = message.uMsg;
			TooltipMessage.wParam = message.wParam;
			TooltipMessage.lParam = message.lParam;
			SendMessage(wharfHints, TTM_RELAYEVENT, 0, (LPARAM)&TooltipMessage);
		}
	}

	if (message.wParam & MK_LBUTTON)
	{
		for (int x = 0; x < wharfs.size(); x++)  // could be more efficent?
		{
			if ((x == i) && wharfs[x]->m_bPressed)
				wharfs[i]->togglePressOffset(true);
			else if (!wharfs[x]->m_bSubwharfOpen)
				wharfs[x]->togglePressOffset(false);
		}
	}
}


void MainWnd_Activate(Message& message)
{
	if (wharfCloseOnSwitch && wharfs.size() && message.wParamLo == WA_INACTIVE)
	{
		if (!PartOfWharf((HWND)(message.lParam)))
		{
			CloseAllFolders();
		}
	}
}


LRESULT CALLBACK MainWndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	Message message;

	message.uMsg = uMsg;
	message.wParam = wParam;
	message.lParam = lParam;
	message.lResult = 0;

	switch (uMsg)
	{
	case LM_GETREVID:
		MainWnd_GetRevId(message);
		break;
	case WM_ENDSESSION:
	case WM_QUERYENDSESSION:
		return SendMessage(hwndLitestep, message.uMsg, message.wParam, message.lParam);
	case WM_SYSCOMMAND:
		if (wParam == SC_CLOSE)
			PostMessage(hwndLitestep, WM_KEYDOWN, LM_SHUTDOWN, 0);
		else
			return DefWindowProc(hWnd, message.uMsg, message.wParam, message.lParam);
		break;
	case LM_RECYCLE:
		SendMessage(hwndLitestep, message.uMsg, message.wParam, message.lParam);
		break;
	case LM_SHADETOGGLE:
		MainWnd_ShadeToggle(message);
		break;
	case WM_TIMER:
		MainWnd_Timer(hWnd, message);
		break;
	case WM_NCLBUTTONDOWN:
	case WM_NCRBUTTONDOWN:
		MainWnd_NCButtonDown(hWnd, message);
		break;
	case WM_NCLBUTTONDBLCLK:
		MainWnd_NCLButtonDblClick(hWnd, message);
		break;
	case WM_WINDOWPOSCHANGING:
		MainWnd_WindowPosChanging(hWnd, message);
		break;
	case WM_NCHITTEST:
		MainWnd_NCHitTest(hWnd, message);
		break;
	case WM_MOVE:
		MainWnd_Move(hWnd, message);
		break;
	case WM_CREATE:
	case WM_ERASEBKGND:
		break;
	case WM_DISPLAYCHANGE:
		MainWnd_DisplayChange(message);
		break;
	case WM_PAINT:
		MainWnd_Paint(hWnd, message);
		break;
	case WM_LBUTTONDOWN:
		MainWnd_LButtonDown(hWnd, message);
		break;
	case WM_LBUTTONUP:
		MainWnd_LButtonUp(hWnd, message);
		break;
	case WM_MOUSEMOVE:
		MainWnd_MouseMove(hWnd, message);
		break;
	case WM_ACTIVATE:
		MainWnd_Activate(message);
		break;
	case WM_RBUTTONDOWN:
	{
		POINT point = {message.lParamLo, message.lParamHi};
		ClientToScreen(hWnd, &point);
		SendMessage(GetLitestepWnd(), LM_POPUP, 0, MAKELPARAM(point.x, point.y));
		break;
	}
	case LM_REGISTERMESSAGE:
	case LM_UNREGISTERMESSAGE:
		SendMessage(GetLitestepWnd(), uMsg, wParam, lParam);
		break;
	default:
		return DefWindowProc(hWnd, uMsg, wParam, lParam);
	}
	return message.lResult;
}


bool PartOfWharf(HWND myWnd)
{
	int i, j;
	if (myWnd == NULL)
		return false;
	if (hMainWnd == myWnd)
		return true;
	if (HiddenWnd == myWnd)
		return TRUE;
	if (hwndLitestep == myWnd)
		return true;

	for (i = 0;i < wharfs.size();i++)
	{
		if (wharfs[i]->m_hWnd == myWnd)
			return true;
	}
	for (i = 0;i < wharfs.size(); i++)
	{
		if (wharfs[i]->m_hwndSubwharf == myWnd)
			return true;
	}

	for (i = 0; i < wharfs.size(); i++)
	{
		for (j = 0; j < wharfs[i]->subwharfs.size(); j++)
		{
			if (myWnd == wharfs[i]->subwharfs[j]->m_hWnd ||
				myWnd == wharfs[i]->subwharfs[j]->taskWnd)
			{
				return true;
			}
		}
	}
	return false;
}

void ShowWharf(void)
{
	ShowWindow(HiddenWnd, SW_HIDE);
	ShowWindow(hMainWnd, SW_SHOWNORMAL);
	wharfHidden = FALSE;
}

void HideWharf(void)
{
	ShowWindow(HiddenWnd, SW_SHOWNORMAL);
	ShowWindow(hMainWnd, SW_HIDE);
	wharfHidden = TRUE;
}

void SetWorkArea(void)
{
	RECT deskRect;
	int oldDocked = wharfDocked;

	SystemParametersInfo(SPI_GETWORKAREA, 0, &deskRect, 0);

	if (wharfBaseX == 0)
	{
		wharfDocked = 1;

		if (setDesktopArea == FALSE)
		{
			deskRect.left = (wharfAutoHide) ? hiddenWidth : wharfTileSize;
			deskRect.right = ScreenWidth;
		}
	}
	else if (wharfBaseX + wharfTileSize == ScreenWidth)
	{
		wharfDocked = 2;

		if (setDesktopArea == FALSE)
		{
			deskRect.left = 0;
			deskRect.right = (wharfAutoHide) ? ScreenWidth - hiddenWidth : ScreenWidth - wharfTileSize;
		}
	}
	else
	{
		wharfDocked = 0;

		if (setDesktopArea == FALSE)
		{
			deskRect.left = 0;
			deskRect.right = ScreenWidth;
		}
	}

	if (wharfDocked != oldDocked)
	{
		SystemParametersInfo(SPI_SETWORKAREA, (isFirst ? 0 : 1), &deskRect, SPIF_SENDCHANGE);
	}

	if (wharfAutoHide && wharfDocked)
	{
		switch (wharfDocked)
		{
			case 1:
			SetWindowPos(HiddenWnd, 0, 0, wharfBaseY, 0, 0, SWP_NOACTIVATE | SWP_NOSENDCHANGING | SWP_NOSIZE | SWP_NOZORDER );
			break;
			case 2:
			SetWindowPos(HiddenWnd, 0, ScreenWidth - hiddenWidth, wharfBaseY, 0, 0, SWP_NOACTIVATE | SWP_NOSENDCHANGING | SWP_NOSIZE | SWP_NOZORDER );
			break;
		}
	}
}


void wharfType::togglePressOffset(bool bToggle)
{
	int i = GetWharfByPointer(this);
	if (wharfDirection == WHARF_DOWN)
	{
		if (m_bTogglePressed != bToggle) // redundancy check...see below comment
			SetWindowPos(m_hWnd,
			             0,
			             bToggle ? wharfPressOffset : 0,
			             bToggle ? (wharfTileSize * i) + titleSize + wharfPressOffset : (wharfTileSize * i) + titleSize,
			             wharfTileSize,
			             bToggle ? wharfTileSize - 1 : wharfTileSize,
			             SWP_NOZORDER);
	}
	else if (wharfDirection == WHARF_RIGHT)
	{
		if (m_bTogglePressed != bToggle) // redundancy check...see below comment
			SetWindowPos(m_hWnd,
			             0,
			             bToggle ? (wharfTileSize*i) + wharfPressOffset : (wharfTileSize*i),
			             bToggle ? wharfPressOffset : 0,
			             bToggle ? (wharfTileSize - 1) : wharfTileSize,
			             wharfTileSize,
			             SWP_NOZORDER);
	}
	// could this be replaced by checking the windows current position for a 1, 1 offset?
	m_bTogglePressed = bToggle;
}

// -------------------------------------------------------------------------------------------------------
// cleanup (opposite of init()). Destroys the window, unregisters the window class

void quitModule(HINSTANCE dllInst)
{
	RECT deskRect;
	int i;

	if (oldCapt)
	{
		SetCapture(oldCapt);
		oldCapt = NULL;
	}
	if (!wharfNoHints)
	{
		if (wharfHints)
		{
			DestroyWindow (wharfHints);
			wharfHints = NULL;
		}
	}
	for (i = 0; i < wharfs.size(); i++)
	{
		wharfs[i]->Destroy();
		delete wharfs[i];
	}
	wharfs.clear();
	SendMessage(hwndLitestep, LM_UNREGISTERMESSAGE, (WPARAM) hMainWnd, (LPARAM) Msgs);

	if (hMainRgn)
	{
		DeleteObject(hMainRgn);
	}
	DestroyWindow(hMainWnd);
	if (wharfAutoHide)
	{
		DestroyWindow(HiddenWnd);
	}

	if (defaultBackImage)
	{
		DeleteObject(defaultBackImage);
	}
	if (folderImage)
	{
		DeleteObject(folderImage);
	}
	if (defaultFolderImage)
	{
		DeleteObject(defaultFolderImage);
	}
	if (titlebarImage)
	{
		DeleteObject(titlebarImage);
	}

	// Reset the work area since we no longer have a docked wharf, or any wharf at all, hah!
	SystemParametersInfo(SPI_GETWORKAREA, 0, &deskRect, 0);
	deskRect.left = 0;
	deskRect.right = ScreenWidth;
	SystemParametersInfo(SPI_SETWORKAREA, 0, &deskRect, 0);

	UnregisterClass(szMainWindowClass, dllInst); // unregister window class
	if (wharfAutoHide)
		UnregisterClass(szHiddenWindowClass, dllInst);

	RemoveBangCommand("!WharfRollup");
    RemoveBangCommand("!WharfShow");
    RemoveBangCommand("!WharfHide");
}

void ReadConfig (void)
{
	char szBuf[MAX_LINE_LENGTH], *tmp;
	char szToken[MAX_LINE_LENGTH];

	wharfHide =
	wharfHidden			= GetRCBool("WharfStartRolledUp", TRUE);
    wharfAlwaysOnTop	= GetRCBoolDef("WharfAlwaysOnTop", FALSE);
	wharfSlidersOnTop	= GetRCBoolDef("WharfSlidersOnTop", TRUE);
	setDesktopArea		= GetRCBool("WharfSetDesktopArea", TRUE);
	wharfAutoUnpress	= GetRCBool("WharfAutoUnpress", TRUE);
	wharfAutoClose		= GetRCBool("WharfNoAutoClose", FALSE);
	wharfTitlebar		= GetRCBool("WharfNoTitleBar", FALSE);
	wharfNoAnim			= GetRCBool("WharfNoAnim", TRUE);
	wharfTitles			= GetRCBool("WharfAllTitles", TRUE);

	wharfTileSize		= GetRCInt("WharfTileSize", 64);
	wharfShadeButton	= GetRCInt("WharfShadeButton", (wharfTileSize / 2));

	wharfDblClickXPosition	= GetRCInt("WharfDblClickXPosition", 0);
	wharfDblClickDockOnLeft	= GetRCBool("WharfDblClickDockOnLeft", TRUE);

	capHeight			= GetRCInt("WharfCapHeight", 16);
	nIconSize			= GetRCInt("WharfIconSize",	32);

	titleFore			= GetRCColor("WharfTitleFore", 0xFFFFFF);
	titleBack			= GetRCColor("WharfTitleBack", 0x802020);

	bevelWidth			= GetRCInt("WharfBevelWidth", 1);
	if (bevelWidth < 0)	bevelWidth = 0;
	wharfStep			= GetRCInt("WharfAnimStep", 64);
	wharfAnimTime		= GetRCInt("WharfAnimDelay", 10);

	wharfAutoHide		= GetRCBool("WharfAutoHide", TRUE);

	autoHideDelay		= GetRCInt("WharfAutoHideDelay", 300);
	autoShowDelay		= GetRCInt("WharfAutoShowDelay", autoHideDelay);
	hiddenWidth			= GetRCInt("WharfHiddenWidth", 1);
	wharfPressOffset	= GetRCInt("WharfPressOffset", 1);

	wharfNoHints		= GetRCBool("WharfNoHints", TRUE);
	wharfCloseOnSwitch	= GetRCBool("WharfCloseOnSwitch", TRUE);

	GetRCString("WharfOpenSound",	wharfOpenSound,	"", MAX_LINE_LENGTH);
	GetRCString("WharfCloseSound",	wharfCloseSound,"", MAX_LINE_LENGTH);
	GetRCString("WharfMinSound",	wharfMinSound,	"", MAX_LINE_LENGTH);
	GetRCString("WharfMaxSound",	wharfMaxSound,	"", MAX_LINE_LENGTH);

    GetRCLine("WharfTitlebarLeftClick",	szTitleBarLeftClick,	MAX_LINE_LENGTH, "!WharfRollup");
    GetRCLine("WharfTitlebarRightClick",szTitleBarRightClick,	MAX_LINE_LENGTH, "");

	snapTo				= GetRCBool("WharfSnapToEdges", TRUE);
	snapToSensitivity	= GetRCInt("WharfSnapSensitivity", 16);

	LSGetImagePath(szImagePath, MAX_LINE_LENGTH);

	if (GetRCString("WharfDefaultBackPix", szDefPixmap, "", MAX_LINE_LENGTH))
	{
		defaultBackImage = LoadLSImage (szDefPixmap, NULL);
	}

	if (GetRCString("WharfFolderBackPix", szFolderPixmap, "", MAX_LINE_LENGTH))
	{
		defaultFolderImage = LoadLSImage (szFolderPixmap, NULL);
	}

	if (GetRCString("WharfFolderPix", szToken, "", MAX_LINE_LENGTH))
	{
		folderImage = LoadLSImage (szToken, NULL);
	}

	if (GetRCString("WharfTitlebarPix", szToken, "", MAX_LINE_LENGTH))
	{
		titlebarImage = LoadLSImage (szToken, NULL);
	}

	GetRCString("ModulesINI", szBuf, "$LiteStepDir$modules.ini", MAX_LINE_LENGTH);
	GetRCString("WharfModulesINI", szBuf, szBuf, MAX_LINE_LENGTH);
	VarExpansionEx(szModulesINI, szBuf, MAX_LINE_LENGTH);

	GetPrivateProfileString("Wharf", "Position", "-64,0", (char *)&szBuf, sizeof(szBuf), szModulesINI);
	tmp = strtok(szBuf, ",");
	wharfBaseX = atoi(tmp);
	tmp = strtok(NULL, "");
	wharfBaseY = atoi(tmp);
	if (wharfBaseX < 0)
		wharfBaseX += ScreenWidth;
	if (wharfBaseY < 0)
		wharfBaseY += ScreenHeight;

	GetRCString("WharfTitleFont", szTitleFont, "Arial", MAX_LINE_LENGTH);
	iTitleFontSize = GetRCInt("WharfTitleFontSize", 8);
	iTitleFontWeight = GetRCInt("WharfTitleFontWeight", 400);
}

void CloseAllFolders()
{
	__try
	{
	    for (int i = 0; i < wharfs.size(); i++)
		{
			if (wharfs[i]->m_bSubwharfOpen || wharfs[i]->m_bMoving)
		    {
			    wharfs[i]->Close();
		    }
	    }
	}
	__except(GetExceptionCode() == EXCEPTION_ACCESS_VIOLATION)
	{
	}
}

void GetWharfData(wharfDataType *wd)
{
	StringCchCopy(lsPath, MAX_LINE_LENGTH, szLitestepPath);
	if (lsPath[strlen(lsPath)] != '\\')
	{
		StringCchCat(lsPath, MAX_LINE_LENGTH, "\\");
	}

	wd->defaultBmp		= szDefPixmap;
	wd->lsPath			= lsPath;
	wd->borderSize		= bevelWidth;
}

void wharfType::Execute()
{
	if (strcmpi(m_szCommand, "Folder") == 0)
	{
		if (CloseFolders()) return;
		ExecuteFolder();
	}
	else if (m_szCommand[0] == '!')
	{
		if (!strcmpi(m_szCommand, "!WharfTasks"))
			ExecuteTasks();
		else
			ParseBangCommand(m_hWnd, m_szCommand, m_szParameters);
	}
	else if (m_szCommand[0] == '@')
	{
		wharfDataType wharfData;
		ZeroMemory(&wharfData, sizeof (wharfData));
		GetWharfData(&wharfData);

		m_pInitWharfModule(m_hWnd, m_hInst, &wharfData);
	}
	else
	{
		if (!wharfAutoUnpress)
		{
			m_bPressed = false;
			togglePressOffset(false);
		}

		ExecuteCommand(NULL);
	}
}


BOOL CALLBACK WharfTasksEnumWindowsProc(HWND hwnd, LPARAM lParam)
{
	WindowVector *v = (WindowVector*)lParam;
	if (IsAppWindow(hwnd)) v->push_back(hwnd);
	return TRUE;
}

void wharfType::ExecuteTasks()
{
	WindowVector windowList;
	char capt[MAX_LINE_LENGTH];

	if (CloseFolders()) return;

	if (m_bMoving || m_bSubwharfOpen)
	{
		ExecuteFolder();
		return;
	}

	ClearFolder();

	m_bSubwharfOpen = false;

	if (!m_hwndSubwharf)
	{
		m_hwndSubwharf = CreateWindowEx(
			(wharfSlidersOnTop ? WS_EX_TOPMOST : 0) | WS_EX_TOOLWINDOW,
			szMainWindowClass,
			"LSWharflet",
			WS_CLIPCHILDREN | WS_POPUP,
			0, 0,
			0, 0,
			wharfSlidersOnTop ? NULL : hwndLitestep,
			NULL,
			dll,
			NULL
			);
	}

	SetWindowLong(m_hwndSubwharf, GWL_USERDATA, magicDWord);
	ShowWindow(m_hwndSubwharf, SW_SHOWNORMAL);

	EnumWindows(WharfTasksEnumWindowsProc, (LPARAM)&windowList);

	for (int n = 0; n < windowList.size(); n++)
	{
		HICON hIcon = NULL;
		HWND parent, window;

		window = windowList[n];

		if (!IsAppWindow(window)) continue;

		subwharfType *shortcut = new subwharfType(this);
		subwharfs.push_back(shortcut);

		capt[0] = 0;
		GetWindowText(window, (char *)&capt, sizeof(capt));
		StringCchCopy(shortcut->m_szName, MAX_LINE_LENGTH, capt);
		StringCchCopy(shortcut->m_szCommand, MAX_LINE_LENGTH, "!none");

		hIcon = GetIconFromWindow(window, true);
		if (hIcon)
			shortcut->m_hIcon = hIcon;
		else
		{
			shortcut->m_hIcon = LoadIcon(NULL, IDI_APPLICATION);
		}
		shortcut->taskWnd = window;
		shortcut->m_bShowBackground = true;
		shortcut->Create("");
	}

	windowList.clear();

	SetWindowRgn(m_hwndSubwharf, m_hrgnSubRgn, TRUE);	

	Open();
}

void CreateWharfhints(HWND hWnd, char *txt, RECT *r)
{
	TOOLINFO ti;    // tool information

	if (wharfNoHints) return;

	ti.cbSize = sizeof(TOOLINFO);
	ti.uFlags = TTF_SUBCLASS;
	ti.hwnd = hWnd;
	ti.hinst = dll;
	ti.uId = 0;
	ti.lpszText = txt;
	ti.rect = *r;

	SendMessage(wharfHints, TTM_ADDTOOL, 0, (LPARAM) (LPTOOLINFO) &ti);
}

void RemoveWharfhints(HWND hWnd)
{
	TOOLINFO ti;    // tool information
	RECT r = {0, 0, 0, 0};

	ti.cbSize = sizeof(TOOLINFO);
	ti.uFlags = 0;
	ti.hwnd = hWnd;
	ti.hinst = dll;
	ti.uId = 0;
	ti.lpszText = NULL;
	ti.rect = r;

	SendMessage(wharfHints, TTM_DELTOOL, 0, (LPARAM) (LPTOOLINFO) &ti);
}

wharfType::wharfType()
{
	m_hInst = NULL;
	m_hWnd = NULL;
	m_hwndSubwharf = NULL;
	m_hbmBackImage = NULL;
	m_hbmFrontImage = NULL;
	m_hIcon = NULL;
	m_szName[0] = 0;
	m_szCommand[0] = 0;
	m_szParameters[0] = 0;
	m_bSubwharfOpen = false;
	m_bMoving = false;
	m_bClosing = false;
	m_bShowBackground = true;
	m_bPressed = false;
	m_bTogglePressed = false;
	m_nMovingPos = 0;
	m_pInitWharfModule = NULL;
	m_pQuitWharfModule = NULL;
	m_pGetLSRegion = NULL;
	m_hrgnSubRgn = NULL;
	m_pDropTarget = NULL;
}

void wharfType::Paint(HDC hdc, HDC src)
{
	HBITMAP oldBMP = NULL;

	// have a background image?
	if (m_hbmBackImage)
	{
		oldBMP = (HBITMAP)SelectObject(src, m_hbmBackImage);
		// are we pressed down?
		if (m_bPressed)
			TransparentBltLS(hdc, 0, 0, wharfTileSize, wharfTileSize, src, 0, 0, RGB(255, 0, 255));
		else
			BitBlt(hdc, 0, 0, wharfTileSize, wharfTileSize, src, 0, 0, SRCCOPY);
	}

	// displaying titles, but not a wharf module?
	if (wharfTitles && !m_hInst)
	{
		PaintTitle(hdc, src);
	}
	if (bevelWidth) // displaying a beveled edge?
	{
		RECT r = {0, 0, wharfTileSize, wharfTileSize};
		Frame3D(hdc, r, RGB(255, 255, 255), RGB(0, 0, 0), bevelWidth);
	}

	if (oldBMP)
	{
		SelectObject(src, oldBMP);
	}
}

void wharfType::PaintTitle(HDC hdc, HDC src)
{
	RECT r;
	HBRUSH back;
	HFONT titleFont, oldFont;
	int nHeight;
	int len = lstrlen(m_szName);

	back = CreateSolidBrush(titleBack);
	GetClientRect(m_hWnd, &r);
	nHeight = MulDiv(iTitleFontSize, GetDeviceCaps(src, LOGPIXELSY), 72);
	r.bottom = nHeight + 2;
	FillRect(src, &r, back);
	titleFont = CreateFont(
	                nHeight,  // 8,
	                0,  // 6,
	                0,
	                0,
	                iTitleFontWeight,  // 0,
	                FALSE,
	                FALSE,
	                FALSE,
	                DEFAULT_CHARSET,
	                OUT_DEFAULT_PRECIS,
	                CLIP_DEFAULT_PRECIS,
	                DEFAULT_QUALITY,
	                DEFAULT_PITCH | FF_DONTCARE,  // DEFAULT_PITCH,
	                szTitleFont
	            );
	oldFont = (HFONT)SelectObject(src, titleFont);
	SetBkMode(src, TRANSPARENT);
	SetTextColor(src, titleFore);
	TextOut(src, 2, 1, m_szName, len);
	BitBlt(hdc, 0, 0, wharfTileSize, wharfTileSize, src, 0, 0, SRCCOPY);
	SelectObject(src, oldFont);
	DeleteObject(titleFont);
	DeleteObject(back);
}

subwharfType::subwharfType(wharfType *p) : wharfType()
{
	parent = p;
	taskWnd = 0;
}

void subwharfType::Execute()
{
	if (m_szCommand[0] == '!')
	{
		if (strcmpi(parent->m_szCommand, "!WharfTasks") != 0)
			ParseBangCommand(parent->m_hwndSubwharf, m_szCommand, m_szParameters);
		else
		{
			if (!SendMessage(hwndLitestep, LM_BRINGTOFRONT, 0, (long)taskWnd))
			{
				SwitchToThisWindow(taskWnd, 1);
			}
		}
	}
	else if (m_szCommand[0] == '@')
	{
		wharfDataType wharfData;
		ZeroMemory(&wharfData, sizeof (wharfData));
		GetWharfData(&wharfData);

		m_pInitWharfModule(m_hWnd, m_hInst, &wharfData);
	}
	else
	{
		ExecuteCommand(NULL);
	}
}

inline void SoundOpen(void)
{
	sndPlaySound(wharfOpenSound, SND_ASYNC | SND_NODEFAULT);
}

inline void SoundClose(void)
{
	sndPlaySound(wharfCloseSound, SND_ASYNC | SND_NODEFAULT);
}

inline void SoundMin(void)
{
	sndPlaySound(wharfMinSound, SND_ASYNC | SND_NODEFAULT);
}

inline void SoundMax(void)
{
	sndPlaySound(wharfMaxSound, SND_ASYNC | SND_NODEFAULT);
}


void BangWharfRollup(HWND hCaller, LPCSTR /* pszArgs */)
{
    SendMessage(hMainWnd, LM_SHADETOGGLE, 0, 0);
}

void BangWharfHide(HWND hCaller, LPCSTR /* pszArgs */)
{
    if (IsWindowVisible(hMainWnd))
    {
        ShowWindow(hMainWnd, SW_HIDE);
    }
}

void BangWharfShow(HWND hCaller, LPCSTR /* pszArgs */)
{
    if (!IsWindowVisible(hMainWnd))
    {
        ShowWindow(hMainWnd, SW_SHOW);
    }
}

void wharfType::Open()
{
	if (wharfOpenSound[0]) SoundOpen();
	m_bSubwharfOpen = true;
	if (wharfNoAnim)
	{
		ApplyFolderPosition(subwharfs.size() * wharfTileSize);
		m_bSubwharfOpen = true;
	}
	else
	{
		m_bMoving = true;
		m_bClosing = false;
		SetTimer(m_hwndSubwharf, 2, wharfAnimTime, NULL);
		OpenFolders++;
	}
}


void wharfType::Close()
{
	m_bPressed = false;
	togglePressOffset(false);
	if (wharfNoAnim)
	{
		SetWindowPos(m_hwndSubwharf, 0, 0, 0, 0, 0, SWP_NOZORDER);
	}
	else
	{
		m_bMoving= true;
		m_bClosing = true;
		SetTimer(m_hwndSubwharf, 3, wharfAnimTime, NULL);
	}
	m_bSubwharfOpen = false;
	OpenFolders--;
}


void wharfType::onTimerOpen()
{
	if (m_bClosing)
	{
		KillTimer(m_hwndSubwharf, 2);
		SetTimer(m_hwndSubwharf, 3, wharfAnimTime, NULL);
	}
	else
	{
		if (m_nMovingPos + wharfStep >= subwharfs.size() * wharfTileSize)
			m_nMovingPos = subwharfs.size() * wharfTileSize;
		else
			m_nMovingPos += wharfStep;
		KillTimer(m_hWnd,  4);

		ApplyFolderPosition(m_nMovingPos);

		if (m_nMovingPos >= subwharfs.size() * wharfTileSize)
		{
			m_bSubwharfOpen = true;
			m_bMoving = false;
			m_bClosing = true;
			KillTimer(m_hwndSubwharf, 2);
		}
	}
}


void wharfType::onTimerClose()
{
	if (!m_bClosing)
	{
		KillTimer(m_hwndSubwharf, 3);
		SetTimer(m_hwndSubwharf, 2, wharfAnimTime, NULL);
	}
	else
	{
		m_nMovingPos -= wharfStep;
		KillTimer(m_hWnd, 4);
		ApplyFolderPosition(m_nMovingPos);
		if (m_nMovingPos <= 0)
		{
			m_bSubwharfOpen = false;
			m_bPressed = false;
			togglePressOffset(false);
			m_bMoving= false;
			m_bClosing = false;
			KillTimer(m_hwndSubwharf, 3);
		}
	}
}

void wharfType::ExecuteFolder()
{
	if (m_bMoving)
	{
		if (m_bClosing)
		{
			m_bClosing = false;
			if (wharfOpenSound[0]) SoundOpen();
		}
		else
		{
			m_bClosing = true;
			if (wharfCloseSound[0]) SoundClose();
		}
	}
	else if (!m_bSubwharfOpen)
	{
		if (wharfCloseSound[0]) SoundOpen();
		Open();
	}
	else if (m_bSubwharfOpen)
	{
		if (wharfCloseSound[0]) SoundClose();
		Close();
	}
}

bool wharfType::CloseFolders()
{
	if (wharfClosing)
	{
		CloseAllFolders();
		return true;
	}

	// Below closes all folders if WharfAutoClose is actiavted...
	if (wharfAutoClose)
	{
		for (int q = 0; q < wharfs.size(); q++)
		{
			if ((wharfs[q]->m_bSubwharfOpen || wharfs[q]->m_bMoving) && wharfs[q] != this)
			{
				wharfs[q]->Close();
			}
		}
	}
	return false;
}

void wharfType::ApplyFolderPosition(int nPos)
{
	int i = GetWharfByPointer(this);
	if ((wharfBaseX + (wharfTileSize / 2)) > ScreenWidth / 2)
	{
		SetWindowPos(
			m_hwndSubwharf,
			0,
			wharfBaseX - nPos,
			wharfBaseY + (i * wharfTileSize) + titleSize,
			nPos,
			wharfTileSize,
			SWP_NOZORDER);
	}
	else
	{
		SetWindowPos(
			m_hwndSubwharf,
			0,
			wharfBaseX + wharfTileSize,
			wharfBaseY + (i * wharfTileSize) + titleSize,
			nPos,
			wharfTileSize,
			SWP_NOZORDER);
	}
}


void wharfType::LoadModule()
{
	INITWHARFMODULEPROC FinitWharfModule;
	QUITWHARFMODULEPROC FquitWharfModule;
	GETLSREGIONPROC FGetLSRegion;

	char *module;
	HINSTANCE moduleInst;

	module = m_szCommand;
	if (*module == '@') module++;
	moduleInst = LoadLibrary(module);
	if ((UINT)moduleInst > HINSTANCE_ERROR)
	{
		FinitWharfModule = (INITWHARFMODULEPROC)GetProcAddress(moduleInst, "initWharfModule");
		if (!FinitWharfModule)
		{
			FinitWharfModule = (INITWHARFMODULEPROC)GetProcAddress(moduleInst, "_initWharfModule");
		}
		FquitWharfModule = (QUITWHARFMODULEPROC)GetProcAddress(moduleInst, "quitWharfModule");
		if (!FquitWharfModule)
		{
			FquitWharfModule = (QUITWHARFMODULEPROC)GetProcAddress(moduleInst, "_quitWharfModule");
		}
		FGetLSRegion = (GETLSREGIONPROC)GetProcAddress(moduleInst, "GetLSRegion");
		if (!FGetLSRegion)
		{
			FGetLSRegion = (GETLSREGIONPROC)GetProcAddress(moduleInst, "_GetLSRegion");
		}

		if (!FinitWharfModule || !FquitWharfModule)
		{
			char errmsg[MAX_LINE_LENGTH];
			StringCchPrintf(errmsg, MAX_LINE_LENGTH, "Invaild Wharf Module:\n%s", module);
			ErrorBox(errmsg);
			FreeLibrary(moduleInst);
			m_hInst = NULL;
		}
		else
		{
			m_hInst = moduleInst;
			m_pInitWharfModule = FinitWharfModule;
			m_pQuitWharfModule = FquitWharfModule;
			m_pGetLSRegion = FGetLSRegion;
			Execute();
		}
	}
	else
	{
		char errmsg[MAX_LINE_LENGTH];
		m_hInst = NULL;
		StringCchPrintf(errmsg, MAX_LINE_LENGTH, "Error loading module:\n%s", module);
		ErrorBox(errmsg);
	}
}

bool wharfType::CreateTileWindow(HWND hParent, const char *szWindowName)
{
	m_hWnd = CreateWindowEx(
	           WS_EX_TRANSPARENT,
	           szMainWindowClass,
	           szWindowName,
	           WS_CHILD,
	           0, 0,
	           wharfTileSize, wharfTileSize,
	           hParent,
	           NULL,
	           dll,
	           NULL);

	if (!m_hWnd) return false;

	SetWindowLong(m_hWnd, GWL_USERDATA, magicDWord);

	if (m_szCommand[0] != '@' && stricmp(m_szCommand, "!WharfTasks") != 0)
	{
		m_pDropTarget = new WharfDropTarget;
		m_pDropTarget->Register(this); 
	}

	//Wharf hints
	if (!wharfNoHints)
	{
		RECT r;
		GetClientRect (m_hWnd, &r);
		CreateWharfhints (m_hWnd, m_szName, &r);
	}
	return true;
}

bool subwharfType::Create(const char *szWindowName)
{
	int x = parent->subwharfs.size()-1;

	if (!CreateTileWindow(parent->m_hwndSubwharf, szWindowName))
	{
		RETURN_BOX("Error creating wharf tile", false);
	}

	if (wharfDirection == WHARF_DOWN)
	{
		SetWindowPos(m_hWnd, 0, x * wharfTileSize, 0, 0, 0, SWP_NOZORDER | SWP_NOSIZE);
	}
	if (m_szCommand[0] == '@')
	{
		LoadModule();
	}

	/* Combine our images into one */
	{
		HBITMAP oldBMP1;
		HGDIOBJ defobj;
		HDC src = CreateCompatibleDC(NULL);
		HDC dst = CreateCompatibleDC(NULL);
		HBRUSH hb = CreateSolidBrush(RGB(255, 0, 255));
		RECT r;

		HDC tempDC = GetDC(m_hWnd);
		m_hbmBackImage = CreateCompatibleBitmap(tempDC, wharfTileSize, wharfTileSize);
		ReleaseDC(m_hWnd, tempDC);

		oldBMP1 = (HBITMAP)SelectObject(dst, m_hbmBackImage);
		r.left = 0;
		r.top = 0;
		r.right = wharfTileSize;
		r.bottom = wharfTileSize;
		FillRect(dst, &r, hb);
		if (wharfTitles && !m_hInst)
		{
			HBRUSH hb2 = CreateSolidBrush(RGB(0, 0, 128));
			r.bottom = 10;
			FillRect(dst, &r, hb2);
			DeleteObject(hb2);
		}

		if (defaultFolderImage)
		{
			if (m_bShowBackground)
			{
				defobj = SelectObject(src, defaultFolderImage);
				BitBlt(dst, 0, 0, wharfTileSize, wharfTileSize, src, 0, 0, SRCCOPY);
				SelectObject(src, defobj);
			}
		}

		PaintFrontImage(src, dst);

		DeleteDC(src);
		SelectObject(dst, oldBMP1);
		DeleteDC(dst);
		DeleteObject(hb);
	}
	if (m_hbmBackImage && !m_hInst)
	{
		HRGN wharfRgn = BitmapToRegion(m_hbmBackImage, RGB(255, 0, 255), 0x101010, (wharfTileSize * x), 0);
		if (!parent->m_hrgnSubRgn)
			parent->m_hrgnSubRgn = wharfRgn;
		else
		{
			CombineRgn(parent->m_hrgnSubRgn, parent->m_hrgnSubRgn, wharfRgn, RGN_OR);
			DeleteObject(wharfRgn);
		}
	}
	else
	{
		HRGN rgn;
		if (!m_pGetLSRegion)
		{
			rgn = CreateRectRgn(wharfTileSize * x, 0, wharfTileSize * (x + 1), wharfTileSize);
		}
		else
		{
			HRGN bgRgn = BitmapToRegion(m_hbmBackImage, RGB(255, 0, 255), 0x101010, (wharfTileSize * x), 0);
			rgn = m_pGetLSRegion(wharfTileSize * x, 0);
			if (!rgn)
				rgn = bgRgn;
			else
			{
				CombineRgn(rgn, rgn, bgRgn, RGN_OR);
				DeleteObject(bgRgn);
			}
		}

		if (!parent->m_hrgnSubRgn)
			parent->m_hrgnSubRgn = rgn;
		else
		{
			CombineRgn(parent->m_hrgnSubRgn, parent->m_hrgnSubRgn, rgn, RGN_OR);
			DeleteObject(rgn);
		}
	}

	ShowWindow(m_hWnd, SW_SHOWNORMAL);
	return true;
}


bool wharfType::Create(const char *szWindowName)
{
	int i = GetWharfByPointer(this);
	RECT r;

	if (!CreateTileWindow(hMainWnd, szWindowName))
	{
		RETURN_BOX("Error creating wharf tile", false);
	}

	if (wharfDirection == WHARF_DOWN)
	{
		SetWindowPos(
			hMainWnd,
			0,
			wharfBaseX,
			wharfBaseY,
			wharfTileSize,
			((i + 1)*wharfTileSize) + titleSize,
			SWP_NOZORDER);
		SetWindowPos(
			m_hWnd,
			0,
			0,
			(wharfTileSize * i) + titleSize,
			0,
			0,
			SWP_NOZORDER | SWP_NOSIZE);
	}
	else if (wharfDirection == WHARF_RIGHT)
	{
		SetWindowPos(
			hMainWnd,
			0,
			wharfBaseX,
			wharfBaseY,
			(i + 1) * wharfTileSize,
			wharfTileSize,
			SWP_NOZORDER);
		SetWindowPos(m_hWnd,
			0,
			wharfTileSize * i,
			0,
			0,
			0,
			SWP_NOZORDER | SWP_NOSIZE);
	}

	if (m_szCommand[0] == '@')
	{
		LoadModule();
	}

	/* Combine our images into one */
	{
		HBITMAP oldBMP1;
		HGDIOBJ defobj;
		HDC src = CreateCompatibleDC(NULL);
		HDC dst = CreateCompatibleDC(NULL);
		HBRUSH hb = CreateSolidBrush(RGB(255, 0, 255));
		HDC tempDC = GetDC(m_hWnd);
		RECT r;

		m_hbmBackImage = CreateCompatibleBitmap(tempDC, wharfTileSize, wharfTileSize);
		oldBMP1 = (HBITMAP)SelectObject(dst, m_hbmBackImage);
		r.left = 0;
		r.top = 0;
		r.right = wharfTileSize;
		r.bottom = wharfTileSize;
		FillRect(dst, &r, hb);
		if (wharfTitles && !m_hInst)
		{
			HBRUSH hb2 = CreateSolidBrush(RGB(0, 0, 128));
			r.bottom = 10;
			FillRect(dst, &r, hb2);
			DeleteObject(hb2);
		}
		if (m_hInst && !m_pGetLSRegion)
		{
			HDC ddc = GetDC(GetDesktopWindow());
			POINT p;

			p.x = 0;
			p.y = 0;
			ClientToScreen(m_hWnd, &p);
			BitBlt(dst, 0, 0, wharfTileSize, wharfTileSize, ddc, p.x, p.y, SRCCOPY);
			ReleaseDC(GetDesktopWindow(), ddc);
		}
		if (defaultBackImage)
		{
			if (m_bShowBackground)
			{
				defobj = SelectObject(src, defaultBackImage);
				TransparentBltLS(dst, 0, 0, wharfTileSize, wharfTileSize, src, 0, 0, RGB(255, 0, 255));
				SelectObject(src, defobj);
			}
		}

		PaintFrontImage(src, dst);

		if	(folderImage && (
							strcmpi(m_szCommand, "Folder") == 0 ||
							strcmpi(m_szCommand, "!WharfTasks") == 0
							)
			)
		{
			if (m_bShowBackground)
			{
				defobj = SelectObject(src, folderImage);
				TransparentBltLS(dst, 0, 0, wharfTileSize, wharfTileSize, src, 0, 0, RGB(255, 0, 255));
				SelectObject(src, defobj);
			}
		}

		ReleaseDC(m_hWnd, tempDC);
		DeleteDC(src);
		SelectObject(dst, oldBMP1);
		DeleteDC(dst);
		DeleteObject(hb);
	}
	/* Calculate transparent regions for hMainWnd here, compile regions per bitmap/tile and add together
	* for a full transparent parent window - <toasty@iav.com>  */
	if (m_hbmBackImage && !m_hInst)
	{
		HRGN wharfRgn;

		wharfRgn = BitmapToRegion(
			m_hbmBackImage,
			RGB(255, 0, 255),
			0x101010,
			0,
			(wharfTileSize * i) + titleSize);

		if (hMainRgn)
			CombineRgn(hMainRgn, hMainRgn, wharfRgn, RGN_OR);
		else
		{
			hMainRgn = CreateRectRgn(0, 0, wharfTileSize, titleSize);
			CombineRgn(hMainRgn, hMainRgn, wharfRgn, RGN_OR);
		}
		DeleteObject(wharfRgn);
	}
	else
	{
		HRGN rgn;
		if (!m_pGetLSRegion)
		{
			rgn = CreateRectRgn(
				0,
				(wharfTileSize * i) + titleSize,
				wharfTileSize,
				(wharfTileSize * i) + wharfTileSize + titleSize);
		}
		else
		{
			HRGN bgRgn = BitmapToRegion(
				m_hbmBackImage,
				RGB(255, 0, 255),
				0x101010,
				0,
				(wharfTileSize * i) + titleSize);
			rgn = m_pGetLSRegion(0, (wharfTileSize * i) + titleSize);
			if (!rgn)
				rgn = bgRgn;
			else
			{
				CombineRgn(rgn, rgn, bgRgn, RGN_OR);
				DeleteObject(bgRgn);
			}
		}
		if (!hMainRgn)
		{
			hMainRgn = rgn;
			rgn = CreateRectRgn(0, 0, wharfTileSize, titleSize);
		}
		CombineRgn(hMainRgn, hMainRgn, rgn, RGN_OR);
		DeleteObject(rgn);
	}

	ShowWindow(m_hWnd, SW_SHOWNORMAL);
	GetClientRect(m_hWnd, &r);
	InvalidateRect(m_hWnd, &r, TRUE);

	return true;
}

void wharfType::LoadImage(const char *szFile, const char *szImage)
{
	// if only '.extract' here
	if (stricmp(".extract", szFile) == 0)
	{
		// get the full file name
		char filename[MAX_LINE_LENGTH];
		StringCchCopy(filename, MAX_LINE_LENGTH, szImage);
		PathFindOnPath(filename, NULL);
		m_hIcon = GetIconFromPathEx(filename, nIconSize, nIconSize);
	}
	else
	{
		if (utils::isIconPath(szFile))
			m_hIcon = LoadLSIcon(szFile, szImage);
		else
			m_hbmFrontImage = LoadLSImage(szFile, szImage);
	}
}


void wharfType::PaintFrontImage(HDC dc, HDC dst)
{
	if (m_hbmFrontImage)
	{
		HGDIOBJ defobj;
		BITMAP bmInfo;

		GetObject(m_hbmFrontImage, sizeof(bmInfo), &bmInfo);
		if (bmInfo.bmWidth > wharfTileSize)
			bmInfo.bmWidth = wharfTileSize;
		if (bmInfo.bmHeight > wharfTileSize)
			bmInfo.bmHeight = wharfTileSize;
		defobj = SelectObject(dc, m_hbmFrontImage);
		TransparentBltLS(
			dst,
			(wharfTileSize / 2) - (bmInfo.bmWidth / 2),
			(wharfTileSize / 2) - (bmInfo.bmHeight / 2),
			bmInfo.bmWidth,
			bmInfo.bmHeight,
			dc,
			0,
			0,
			RGB(255, 0, 255));
		DeleteObject(m_hbmFrontImage);
		SelectObject(dc, defobj);
		m_hbmFrontImage = NULL;
	}
	else if (m_hIcon)
	{
		DrawIconEx(dst,
			(wharfTileSize / 2) - nIconSize/2,
			(wharfTileSize / 2) - nIconSize/2,
			m_hIcon,
			nIconSize,
			nIconSize,
			0,
			NULL,
			DI_NORMAL);
		DeleteObject(m_hIcon);
		m_hIcon = NULL;
	}
}

void wharfType::Destroy()
{
	if (m_pDropTarget)
	{
		m_pDropTarget->Revoke(m_hWnd);
		delete m_pDropTarget;
	}

	ClearFolder();

	if (m_hInst)
	{
		m_pQuitWharfModule(m_hInst);
		FreeLibrary(m_hInst);
	}
	if (m_hwndSubwharf)	DestroyWindow(m_hwndSubwharf);
	if (m_hrgnSubRgn)	DeleteObject(m_hrgnSubRgn);
	DestroyWindow(m_hWnd);
	if (m_hbmBackImage)		DeleteObject(m_hbmBackImage);
	if (m_hbmFrontImage)	DeleteObject(m_hbmFrontImage);
	if (m_hIcon)			DeleteObject(m_hIcon);
}

void wharfType::ClearFolder()
{
	for (int x = 0; x < subwharfs.size(); x++)
	{
		subwharfs[x]->Destroy();
		delete subwharfs[x];
	}
	subwharfs.clear();
}

void wharfType::DropFile(const char *szFile)
{
	if (!m_hwndSubwharf && wharfAutoClose) CloseAllFolders();
	ExecuteCommand(szFile);
}

void wharfType::ExecuteCommand(const char *szExtra)
{
	char szParameters[MAX_LINE_LENGTH];
	char workDirectory[MAX_LINE_LENGTH];
	char drive[_MAX_DRIVE];
	char dir[_MAX_DIR];

	_splitpath(m_szCommand, drive, dir, NULL, NULL);
	StringCchCopy(workDirectory, MAX_LINE_LENGTH, drive);
	StringCchCat(workDirectory, MAX_LINE_LENGTH, dir);

	if (szExtra && szExtra[0])
	{
		if (m_szParameters[0])
		{
			StringCchPrintf(szParameters, MAX_LINE_LENGTH, "%s \"%s\"", m_szParameters, szExtra);
		}
		else
		{
			StringCchPrintf(szParameters, MAX_LINE_LENGTH, "\"%s\"", szExtra);
		}
	}
	else
	{
		StringCchCopy(szParameters, MAX_LINE_LENGTH, m_szParameters);
	}

	LSExecuteEx(NULL, NULL, m_szCommand, szParameters, workDirectory, SW_SHOWNORMAL);
}

void wharfType::SetDropHover()
{
	if (0 == stricmp(m_szCommand, "folder") && !m_bSubwharfOpen)
	{
		if (CloseFolders()) return;
		Open();
	}
}

void wharfType::ClearDropHover()
{

}

