/*
This is a part of the LiteStep Shell Source code.

Copyright (C) 1997-2001 The LiteStep Development Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
#ifndef __WHARF_H
#define __WHARF_H

#include "droptarget.h"	// Added by ClassView
#define WHARF_UP                0
#define WHARF_DOWN              1
#define WHARF_LEFT              2
#define WHARF_RIGHT             3

// forward decl
class subwharfType;
class WharfDropTarget;

typedef std::vector< subwharfType* > SubwharfVector;

typedef int (FAR *INITWHARFMODULEPROC)(HWND, HINSTANCE, wharfDataType*);
typedef int (FAR *QUITWHARFMODULEPROC)(HINSTANCE);
typedef HRGN (FAR *GETLSREGIONPROC)(int, int);

class wharfType
{
public:
	void DropFile(const char *szFile);
	void ClearDropHover();
	void SetDropHover();
	WharfDropTarget *m_pDropTarget;
	void ExecuteCommand(const char *szExtra);
	void ClearFolder();
	void Destroy();
	void PaintFrontImage(HDC dc, HDC dst);
	void LoadImage(const char *szFile, const char *szImage);
	bool CreateTileWindow(HWND hParent, const char *szWindowName);
	void LoadModule();
	void ApplyFolderPosition(int nPos);
	bool CloseFolders();
	void Open();
	void Close();
	void onTimerOpen();
	void onTimerClose();

	bool Create(const char *szWindowName);
	void Execute();
	void ExecuteTasks();
	void ExecuteFolder();
	void togglePressOffset(bool bToggle);

	wharfType();

	virtual void Paint(HDC hdc, HDC src);
	virtual void PaintTitle(HDC hdc, HDC src);

	HINSTANCE	m_hInst;
	HWND		m_hWnd;
	HWND		m_hwndSubwharf;

	HBITMAP	m_hbmBackImage;
	HBITMAP	m_hbmFrontImage;
	HICON	m_hIcon;
	bool	m_bBitmap;

	char m_szName[MAX_LINE_LENGTH];
	char m_szCommand[MAX_LINE_LENGTH];
	char m_szParameters[MAX_LINE_LENGTH];
	bool m_bSubwharfOpen;
	bool m_bMoving;
	bool m_bClosing;
	bool m_bShowBackground;
	bool m_bPressed;
	bool m_bTogglePressed;
	int m_nMovingPos;
	INITWHARFMODULEPROC	m_pInitWharfModule;
	QUITWHARFMODULEPROC	m_pQuitWharfModule;
	GETLSREGIONPROC		m_pGetLSRegion;
	HRGN m_hrgnSubRgn;

	SubwharfVector subwharfs;
};

class subwharfType : public wharfType
{
	wharfType *parent;

public:
	bool Create(const char* szWindowName);
	subwharfType(wharfType *p);

	void Execute();

	HWND taskWnd;
};

typedef std::vector< wharfType* > WharfVector;
typedef std::vector<HWND> WindowVector;

extern "C"
{
	__declspec( dllexport ) int initModuleEx(HWND parent, HINSTANCE dll, LPCSTR szPath);
	__declspec( dllexport ) void quitModule(HINSTANCE dllInst);
}

#endif

